@@ -125,6 +125,7 @@
 extern crate alloc;
 use alloc::alloc::Layout;
 use alloc::borrow::{Cow, ToOwned};
+use alloc::boxed::Box;
 use alloc::string::String;
 use alloc::vec::Vec;
 use core::any::Any;
@@ -261,6 +262,21 @@ impl From<Infallible> for AllocError {
 	}
 }
 
+/// Reimplementation of `ptr::set_ptr_value` as long as that one is unstable
+///
+/// Constructs a new pointer to `addr_ptr` with the metadata and type of `meta_ptr`.
+#[inline]
+fn set_ptr_value<T: ?Sized, U>(mut meta_ptr: *const T, addr_ptr: *mut U) -> *mut T {
+	let thin = (&mut meta_ptr as *mut *const T).cast::<*const u8>();
+	// Safety: In case of a thin pointer, this operations is identical
+	// to a simple assignment. In case of a fat pointer, with the current
+	// fat pointer layout implementation, the first field of such a
+	// pointer is always the data pointer, which is likewise assigned.
+	unsafe { *thin = addr_ptr.cast() };
+
+	meta_ptr as *mut T
+}
+
 /// Metadata part of a shared allocation.
 struct RcMeta {
 	/// Id for the thread which may use local references
@@ -300,6 +316,88 @@ impl<T: ?Sized> RcBox<T> {
 		let layout = Layout::for_value(unsafe { ptr.as_ref() });
 		unsafe { alloc::alloc::dealloc(ptr.as_ptr().cast(), layout) };
 	}
+
+	/// Tries to allocate an `RcBox` for a possibly dynamically sized value
+	///
+	/// Size and alignment of `example` are used for allocation and if `example` is a fat reference
+	/// the pointer metadata is copied to the resulting pointer.
+	///
+	/// Returns a mutable pointer on success and the memory layout that could not be allocated
+	/// if the allocation failed.
+	#[inline]
+	fn try_allocate_for_val(
+		meta: RcMeta,
+		example: &T,
+		zeroed: bool,
+	) -> Result<NonNull<RcBox<T>>, Layout> {
+		let layout = Layout::new::<RcBox<()>>();
+		let layout = layout
+			.extend(Layout::for_value(example))
+			.map_err(|_| layout)?
+			.0
+			.pad_to_align();
+
+		// Allocate memory
+		let ptr = unsafe {
+			if zeroed {
+				alloc::alloc::alloc_zeroed(layout)
+			} else {
+				alloc::alloc::alloc(layout)
+			}
+		}
+		.cast::<RcBox<()>>();
+
+		// Write RcMeta fields
+		// Safety: Freshly allocated, so valid to write to.
+		unsafe { ptr::addr_of_mut!((*ptr).meta).write(meta) };
+
+		// Combine metadata from `example` with new memory
+		let result = set_ptr_value(example, ptr);
+
+		NonNull::new(result as *mut RcBox<T>).ok_or(layout)
+	}
+
+	/// Allocates an `RcBox` for a possibly dynamically sized value
+	///
+	/// Size and alignment of `example` are used for allocation and if `example` is a fat reference
+	/// the pointer metadata is copied to the resulting pointer.
+	///
+	/// Returns a mutable pointer on success.
+	///
+	/// # Panics
+	/// Panics or aborts if the allocation failed.
+	#[inline]
+	fn allocate_for_val(meta: RcMeta, example: &T, zeroed: bool) -> NonNull<RcBox<T>> {
+		match Self::try_allocate_for_val(meta, example, zeroed) {
+			Ok(result) => result,
+			Err(layout) => alloc::alloc::handle_alloc_error(layout),
+		}
+	}
+
+	/// Get the pointer to a `RcBox<T>` from a pointer to the data
+	///
+	/// # Safety
+	///
+	/// The pointer must point to (and have valid metadata for) the data part of a previously
+	/// valid instance of `RcBox<T>` and it must not be dangling.
+	#[inline]
+	unsafe fn ptr_from_data_ptr(ptr: *const T) -> *const RcBox<T> {
+		// Calculate layout of RcBox<T> without `data` tail, but including padding
+		let base_layout = Layout::new::<RcBox<()>>();
+		// Safety: covered by the safety contract above
+		let value_alignment = mem::align_of_val(unsafe { &*ptr });
+		let value_offset_layout =
+			Layout::from_size_align(0, value_alignment).expect("invalid memory layout");
+		let layout = base_layout
+			.extend(value_offset_layout)
+			.expect("invalid memory layout")
+			.0;
+
+		// Move pointer to point to the start of the original RcBox<T>
+		// Safety: covered by the safety contract above
+		let rcbox = unsafe { ptr.cast::<u8>().offset(-(layout.size() as isize)) };
+		set_ptr_value(ptr, rcbox as *mut u8) as *const RcBox<T>
+	}
 }
 
 impl<T> RcBox<T> {
@@ -308,7 +406,7 @@ impl<T> RcBox<T> {
 	/// Returns a mutable reference with arbitrary lifetime on success and the memory layout that
 	/// could not be allocated if the allocation failed.
 	#[inline]
-	fn try_allocate<'a>(meta: RcMeta) -> Result<&'a mut RcBox<mem::MaybeUninit<T>>, Layout> {
+	fn try_allocate(meta: RcMeta) -> Result<NonNull<RcBox<mem::MaybeUninit<T>>>, Layout> {
 		let layout = Layout::new::<RcBox<T>>();
 
 		let ptr = unsafe { alloc::alloc::alloc(layout) }.cast::<RcBox<mem::MaybeUninit<T>>>();
@@ -316,7 +414,7 @@ impl<T> RcBox<T> {
 			Err(layout)
 		} else {
 			unsafe { ptr::addr_of_mut!((*ptr).meta).write(meta) };
-			Ok(unsafe { ptr.as_mut().unwrap() })
+			Ok(unsafe { NonNull::new_unchecked(ptr) })
 		}
 	}
 
@@ -327,7 +425,7 @@ impl<T> RcBox<T> {
 	/// # Panics
 	/// Panics or aborts if the allocation failed.
 	#[inline]
-	fn allocate<'a>(meta: RcMeta) -> &'a mut RcBox<mem::MaybeUninit<T>> {
+	fn allocate(meta: RcMeta) -> NonNull<RcBox<mem::MaybeUninit<T>>> {
 		match Self::try_allocate(meta) {
 			Ok(result) => result,
 			Err(layout) => alloc::alloc::handle_alloc_error(layout),
@@ -787,11 +885,44 @@ impl<T: ?Sized, State: RcState> HybridRc<T, State> {
 	pub fn as_ptr(this: &Self) -> *const T {
 		let ptr = this.ptr.as_ptr();
 
-		// SAFETY: Neccessary for `from_raw()` (when implemented), retains provenance.
+		// Safety: Neccessary for `from_raw()` (when implemented), retains provenance.
 		// Besides that, does basically the same thing as `data()` or `get_mut_unchecked()`.
 		unsafe { ptr::addr_of_mut!((*ptr).data) }
 	}
 
+	/// Consumes the `HybridRc<T, State>`, returning the wrapped pointer.
+	///
+	/// To avoid a memory leak the pointer must be converted back to a `HybridRc` using
+	/// [`HybridRc<T, State>::from_raw()`].
+	#[must_use = "Memory will leak if the result is not used"]
+	pub fn into_raw(this: Self) -> *const T {
+		let ptr = Self::as_ptr(&this);
+		mem::forget(this);
+		ptr
+	}
+
+	/// Reconstructs a `HybridRc<T, State>` from a raw pointer.
+	///
+	/// Creates a `HybridRc<T, State>` from a pointer that has been previously returned by
+	/// a call to [`into_raw()`].
+	///
+	/// # Safety
+	///
+	/// The raw pointer must have been previously returned by a call to
+	/// [`HybridRc<T, State>`][`into_raw()`] for the same `State` *and* the same `T` or another
+	/// compatible type that has the same size and alignment. The latter case amounts to
+	/// [`mem::transmute()`] and is likely to produce undefined behaviour if not handled correctly.
+	///
+	/// The value must not have been dropped yet.
+	///
+	/// [`into_raw()`]: Self::into_raw
+	pub unsafe fn from_raw(ptr: *const T) -> Self {
+		// Safety: covered by the safety contract for this function
+		let box_ptr = unsafe { RcBox::<T>::ptr_from_data_ptr(ptr) };
+
+		Self::from_inner(NonNull::new(box_ptr as *mut _).expect("invalid pointer"))
+	}
+
 	/// Creates a new [`Weak`] for the referenced value.
 	///
 	/// # Example
@@ -820,7 +951,7 @@ impl<T: ?Sized, State: RcState> HybridRc<T, State> {
 	/// ```
 	#[inline]
 	pub fn downgrade_pin(this: &Pin<Self>) -> PinWeak<T> {
-		// SAFETY: We are not moving anything and we don't expose a non-pinned pointer.
+		// Safety: We are not moving anything and we don't expose a non-pinned pointer.
 		let this = unsafe { Self::pin_get_ref(this) };
 		PinWeak(Self::downgrade(this))
 	}
@@ -992,7 +1123,8 @@ impl<T, State: RcState> HybridRc<T, State> {
 	/// [`to_local()`]: Self::to_local
 	#[inline]
 	pub fn new(data: T) -> Self {
-		let inner = RcBox::allocate(Self::build_new_meta());
+		let mut inner = RcBox::allocate(Self::build_new_meta());
+		let inner = unsafe { inner.as_mut() };
 		inner.data.write(data);
 		Self::from_inner(unsafe { inner.assume_init() }.into())
 	}
@@ -1001,7 +1133,7 @@ impl<T, State: RcState> HybridRc<T, State> {
 	#[inline]
 	pub fn new_uninit() -> HybridRc<mem::MaybeUninit<T>, State> {
 		let inner = RcBox::allocate(Self::build_new_meta());
-		HybridRc::from_inner(inner.into())
+		HybridRc::from_inner(inner)
 	}
 
 	/// Creates a new `HybridRc` with uninitialized contents, with the memory being filled with
@@ -1012,9 +1144,61 @@ impl<T, State: RcState> HybridRc<T, State> {
 	/// [`MaybeUninit::zeroed()`]: mem::MaybeUninit::zeroed
 	#[inline]
 	pub fn new_zeroed() -> HybridRc<mem::MaybeUninit<T>, State> {
-		let inner = RcBox::allocate(Self::build_new_meta());
-		inner.data = mem::MaybeUninit::zeroed();
-		HybridRc::from_inner(inner.into())
+		let mut inner = RcBox::allocate(Self::build_new_meta());
+		unsafe { inner.as_mut() }.data = mem::MaybeUninit::zeroed();
+		HybridRc::from_inner(inner)
+	}
+
+	/// Creates a new `HybridRc` with a possibly cyclic reference.
+	///
+	/// For this a reference to a [`Weak`] is passed to the closure that – after this function
+	/// returns – will point to the new value itself. Attempting to upgrade the weak reference
+	/// before `new_cyclic` returns will result in a `ValueDropped` error. However, the weak
+	/// reference may be cloned freely and stored for use at a later time.
+	#[inline]
+	pub fn new_cyclic(data_fn: impl FnOnce(&Weak<T>) -> T) -> HybridRc<T, State> {
+		// Construct metadata for an initially non-upgradable RcBox
+		let meta = RcMeta {
+			owner: if State::SHARED {
+				None.into()
+			} else {
+				ThreadId::current_thread().into()
+			},
+			strong_local: Cell::new(0),
+			strong_shared: 0.into(),
+			weak: 1.into(),
+		};
+
+		// Allocate memory (uninitialized)
+		let inner = RcBox::<T>::allocate(meta);
+
+		// Construct `Weak`
+		let weak: Weak<T> = Weak { ptr: NonNull::from(inner).cast() };
+
+		// Run data function, keeping the ownership of the weak reference.
+		let data = data_fn(&weak);
+
+		// Initialize data in our box
+		// Not creating an immediate &mut of the whole box to not invalidate the
+		// weak pointer under Stacked Borrows rules.
+		unsafe { &mut *ptr::addr_of_mut!((*inner.as_ptr()).data) }.write(data);
+
+		// Don't run `Weak`s destructor. The value we just initialized should keep existing and we
+		// need a weak count of 1 for the strong reference that we are currently constructing.
+		mem::forget(weak);
+
+		// Fix the reference counts
+		{
+			let meta = unsafe { &*ptr::addr_of!((*inner.as_ptr()).meta) };
+			if !State::SHARED {
+				meta.inc_strong_local()
+			}
+			// Must be at least `Release`, so that all threads see the initialized data before
+			// they can observe a non-zero reference count.
+			meta.strong_shared.fetch_add(1, Ordering::Release);
+		}
+
+		Self::from_inner(inner.cast())
 	}
 
 	/// Creates a new `Pin<HybridRc<T>>`. If `T` does not implement `Unpin`, then `data` will be
@@ -1033,7 +1217,8 @@ impl<T, State: RcState> HybridRc<T, State> {
 	/// allocation failure happens.
 	#[inline]
 	pub fn try_new(data: T) -> Result<Self, AllocError> {
-		let inner = RcBox::try_allocate(Self::build_new_meta()).map_err(|_| AllocError)?;
+		let mut inner = RcBox::try_allocate(Self::build_new_meta()).map_err(|_| AllocError)?;
+		let inner = unsafe { inner.as_mut() };
 		inner.data.write(data);
 		Ok(Self::from_inner(unsafe { inner.assume_init() }.into()))
 	}
@@ -1065,9 +1250,9 @@ impl<T, State: RcState> HybridRc<T, State> {
 	/// [`MaybeUninit::zeroed()`]: mem::MaybeUninit::zeroed
 	#[inline]
 	pub fn try_new_zeroed() -> Result<HybridRc<mem::MaybeUninit<T>, State>, AllocError> {
-		let inner = RcBox::try_allocate(Self::build_new_meta()).map_err(|_| AllocError)?;
-		inner.data = mem::MaybeUninit::zeroed();
-		Ok(HybridRc::from_inner(inner.into()))
+		let mut inner = RcBox::try_allocate(Self::build_new_meta()).map_err(|_| AllocError)?;
+		unsafe { inner.as_mut() }.data = mem::MaybeUninit::zeroed();
+		Ok(HybridRc::from_inner(inner))
 	}
 
 	/// Returns the inner value, if this `HybridRc` is the only strong reference to it.
@@ -1233,6 +1418,43 @@ impl<T: ?Sized> Rc<T> {
 			Pin::new_unchecked(Self::to_shared(this))
 		}
 	}
+
+	/// Increments the local strong reference count on the `Rc<T>` associated by the given pointer
+	///
+	/// Increases the local strong reference count as if a new `Rc` was cloned and kept alive.
+	/// May panic in the unlikely case the platform-specific maximum for the reference count is
+	/// reached.
+	///
+	/// # Safety
+	/// The pointer must have been obtained through [`HybridRc<T, Local>::into_raw()`], the value
+	/// must still be live and have a local strong count of at least 1 when this method is invoked
+	/// and this call must be performed on the same thread as where the original `Rc` was created.
+	///
+	/// [`HybridRc<T, Local>::into_raw()`]: `Rc::into_raw`
+	#[inline]
+	pub unsafe fn increment_local_strong_count(ptr: *const T) {
+		unsafe {
+			let box_ptr = RcBox::<T>::ptr_from_data_ptr(ptr as *mut T);
+			(*box_ptr).meta.inc_strong_local();
+		}
+	}
+
+	/// Decrements the local strong reference count on the `Rc<T>` associated by the given pointer
+	///
+	/// If the local strong reference counter reaches 0, the value is no longer considered owned
+	/// by the calling thread and if there are no shared strong references to keep the value alive,
+	/// it will be dropped.
+	///
+	/// # Safety
+	/// The pointer must have been obtained through [`HybridRc<T, Local>::into_raw()`], the value
+	/// must still be live and have a local strong count of at least 1 when this method is invoked
+	/// and this call must be performed on the same thread as where the original `Rc` was created.
+	///
+	/// [`HybridRc<T, Local>::into_raw()`]: `Rc::into_raw`
+	#[inline]
+	pub unsafe fn decrement_local_strong_count(ptr: *const T) {
+		mem::drop(unsafe { Rc::from_raw(ptr) });
+	}
 }
 
 impl<T: ?Sized> Arc<T> {
@@ -1309,6 +1531,40 @@ impl<T: ?Sized> Arc<T> {
 			Some(Pin::new_unchecked(Self::to_local(this)?))
 		}
 	}
+
+	/// Increments the shared strong reference count on the `Arc<T>` associated by the given pointer
+	///
+	/// Increases the shared strong reference count as if a new `Arc` was cloned and kept alive.
+	/// May panic in the unlikely case the platform-specific maximum for the reference count is
+	/// reached.
+	///
+	/// # Safety
+	/// The pointer must have been obtained through [`HybridRc<T, Shared>::into_raw()`] and the
+	/// value must still be live when this method is invoked.
+	///
+	/// [`HybridRc<T, Shared>::into_raw()`]: `Arc::into_raw`
+	#[inline]
+	pub unsafe fn increment_shared_strong_count(ptr: *const T) {
+		unsafe {
+			let box_ptr = RcBox::<T>::ptr_from_data_ptr(ptr);
+			(*box_ptr).meta.inc_strong_shared();
+		}
+	}
+
+	/// Decrements the shared strong reference count on the `Arc<T>` associated by the given pointer
+	///
+	/// If the shared strong reference counter (including the implicit shared reference for local
+	/// strong references) reaches 0, the value will be dropped.
+	///
+	/// # Safety
+	/// The pointer must have been obtained through [`HybridRc<T, Shared>::into_raw()`] and the
+	/// value must still be live when this method is invoked.
+	///
+	/// [`HybridRc<T, Shared>::into_raw()`]: `Arc::into_raw`
+	#[inline]
+	pub unsafe fn decrement_shared_strong_count(ptr: *const T) {
+		mem::drop(unsafe { Arc::from_raw(ptr) });
+	}
 }
 
 impl<T: Clone, State: RcState> HybridRc<T, State> {
@@ -1852,6 +2108,30 @@ where
 	}
 }
 
+impl<T: ?Sized, State: RcState> From<Box<T>> for HybridRc<T, State> {
+	#[inline]
+	fn from(src: Box<T>) -> HybridRc<T, State> {
+		let len = mem::size_of_val(&*src);
+		let inner = RcBox::allocate_for_val(Self::build_new_meta(), &*src, false);
+		let dest = unsafe { ptr::addr_of_mut!((*inner.as_ptr()).data) }.cast();
+
+		// Safety: The freshly allocated `RcBox` can't alias `src` and the payload can be fully
+		// moved by copying the memory, because it's not Pin<Box<T>>. `allocate_for_val` ensures
+		// the destination payload buffer is big enough for the value.
+		unsafe {
+			(&*src as *const T)
+				.cast::<u8>()
+				.copy_to_nonoverlapping(dest, len);
+		}
+
+		// Drop original box without running the destructor
+		// Safety: This *should* be sound, as ManuallyDrop<T> has the same layout as T.
+		mem::drop(unsafe { mem::transmute::<Box<T>, Box<mem::ManuallyDrop<T>>>(src) });
+
+		HybridRc::from_inner(inner)
+	}
+}
+
 impl<T, State: RcState> iter::FromIterator<T> for HybridRc<[T], State> {
 	/// Takes each element in the `Iterator` and collects it into an `HybridRc<[T], State>`.
 	///
@@ -2267,6 +2547,15 @@ impl<T: ?Sized> Drop for Weak<T> {
 unsafe impl<T: ?Sized + Sync + Send> Send for Weak<T> {}
 unsafe impl<T: ?Sized + Sync + Send> Sync for Weak<T> {}
 
+/// `PinWeak<T>` represents a non-owning reference to a pinned value managed by a
+/// [`Pin`]`<`[`HybridRc<T, _>`]`>`.
+///
+/// The typical way to obtain a `PinWeak<T>` is to call [`HybridRc::downgrade_pin()`].
+///
+/// See [`Weak<T>`] for more information about weak references.
+///
+/// [`upgrade()`]: PinWeak::upgrade
+/// [`upgrade_local()`]: PinWeak::upgrade_local
 #[repr(transparent)]
 pub struct PinWeak<T: ?Sized>(Weak<T>);
 
@@ -2278,6 +2567,9 @@ impl<T: ?Sized> PinWeak<T> {
 	/// # Errors
 	/// - [`ValueDropped`]: the referenced value has already been dropped.
 	/// - [`WrongThread`]: another thread currently holds `Rc`s for the value.
+	///
+	/// [`ValueDropped`]: UpgradeError::ValueDropped
+	/// [`WrongThread`]: UpgradeError::WrongThread
 	#[inline]
 	pub fn upgrade_local(&self) -> Result<Pin<Rc<T>>, UpgradeError> {
 		Ok(unsafe { Pin::new_unchecked(self.0.upgrade_local()?) })
@@ -2289,6 +2581,8 @@ impl<T: ?Sized> PinWeak<T> {
 	///
 	/// # Errors
 	/// - [`ValueDropped`]: the referenced value has already been dropped.
+	///
+	/// [`ValueDropped`]: UpgradeError::ValueDropped
 	#[inline]
 	pub fn upgrade(&self) -> Result<Pin<Arc<T>>, UpgradeError> {
 		Ok(unsafe { Pin::new_unchecked(self.0.upgrade()?) })
@@ -2297,7 +2591,6 @@ impl<T: ?Sized> PinWeak<T> {
 	/// Gets a lower bound to the number of strong pointers to the inner value.
 	///
 	/// See [`Weak::strong_count()`] for more information.
-	/// ```
 	#[inline]
 	pub fn strong_count(&self) -> usize {
 		self.0.strong_count()
@@ -2310,6 +2603,41 @@ impl<T: ?Sized> PinWeak<T> {
 	pub fn weak_count(&self) -> usize {
 		self.0.weak_count()
 	}
+
+	/// Transforms this `PinWeak<T>` into a [`Weak<T>`]
+	///
+	/// # Safety
+	/// This function is unsafe. You must guarantee that you will continue to treat the `Weak` as
+	/// pinned after you call this function. Not maintaining the pinning invariants that is a
+	/// violation of the API contract and may lead to undefined behavior in later (safe) operations.
+	///
+	/// If the underlying data is [`Unpin`], [`PinWeak::into_inner()`] should be used instead.
+	#[inline]
+	pub unsafe fn into_inner_unchecked(self) -> Weak<T> {
+		self.0
+	}
+}
+
+impl<T> PinWeak<T> {
+	/// Constructs a dummy `PinWeak<T>`, without allocating any memory.
+	///
+	/// Trying to upgrade the result will always result in a [`ValueDropped`] error.
+	///
+	/// [`ValueDropped`]: UpgradeError::ValueDropped
+	pub fn new() -> PinWeak<T> {
+		Self(Weak::new())
+	}
+}
+
+impl<T: ?Sized + Unpin> PinWeak<T> {
+	/// Transforms this `PinWeak<T>` into a [`Weak<T>`]
+	///
+	/// This requires that the data inside the shared allocation is [`Unpin`], so that we
+	/// can ignore the pinning invariants when unwrapping it.
+	#[inline]
+	pub fn into_inner(self) -> Weak<T> {
+		self.0
+	}
 }
 
 impl<T: ?Sized> fmt::Debug for PinWeak<T> {
@@ -2342,3 +2670,17 @@ impl<T: ?Sized> Clone for PinWeak<T> {
 		Self(self.0.clone())
 	}
 }
+
+impl<T> Default for PinWeak<T> {
+	/// Constructs a dummy `PinWeak<T>`, without allocating any memory.
+	///
+	/// See [`PinWeak<T>::new()`].
+	#[inline]
+	fn default() -> Self {
+		Self::new()
+	}
+}
+
+// Safety: Like for Weak<T> T: Send + Sync gives all guarantees we need to make PinWeak Send + Sync.
+unsafe impl<T: ?Sized + Sync + Send> Send for PinWeak<T> {}
+unsafe impl<T: ?Sized + Sync + Send> Sync for PinWeak<T> {}
