@@ -6,6 +6,7 @@ extern crate std;
 use super::*;
 use alloc::{format, vec};
 use core::convert::TryInto;
+use core::fmt::Debug;
 use std::collections::hash_map;
 use std::{thread, thread_local};
 
@@ -471,3 +472,121 @@ fn test_collect() {
 	let a: Rc<[i32]> = vec.into_iter().collect();
 	assert_eq!(&a[..], &[1, 2, 3, 4]);
 }
+
+#[test]
+fn test_from_box() {
+	let b: Box<i64> = Box::new(42);
+	let rc: Rc<i64> = b.into();
+	assert_eq!(*rc, 42);
+}
+
+#[test]
+fn test_from_box_slice() {
+	let vec = vec![1, 2, 3, 4];
+	let b: Box<[i32]> = Box::from(&vec[..]);
+	let rc: Rc<[i32]> = From::<Box<_>>::from(b);
+	assert_eq!(&rc[..], &vec[..]);
+}
+
+#[test]
+fn test_from_box_trait_obj() {
+	let map = hash_map::HashMap::<i32, i64>::new();
+
+	let original_box = Box::new(map.clone());
+	let erased_box: Box<dyn Debug> = original_box;
+	let rc: Rc<dyn Debug> = From::<Box<_>>::from(erased_box);
+	assert_eq!(format!("{:?}", &*rc), format!("{:?}", &map));
+}
+
+#[test]
+fn test_raw_conversion() {
+	let rc: Rc<i64> = Rc::new(1337);
+	let ptr = Rc::into_raw(rc);
+
+	let rc: Rc<i64> = unsafe { Rc::from_raw(ptr) };
+	assert_eq!(*rc, 1337);
+
+	// ---
+
+	let arc: Arc<str> = Arc::from("hallo");
+	let ptr = Arc::into_raw(arc);
+
+	let arc: Arc<str> = unsafe { Arc::from_raw(ptr) };
+	assert_eq!(&*arc, "hallo");
+}
+
+#[test]
+fn test_raw_counter_manipulation() {
+	let rc: Rc<i64> = Rc::new(1337);
+	let weak = Rc::downgrade(&rc);
+	let ptr = Rc::into_raw(rc);
+
+	unsafe {
+		Rc::increment_local_strong_count(ptr);
+		assert!(weak.upgrade().is_ok());
+		Rc::decrement_local_strong_count(ptr);
+		assert!(weak.upgrade().is_ok());
+		Rc::decrement_local_strong_count(ptr);
+		assert!(weak.upgrade().is_err());
+	}
+
+	// ---
+
+	let arc: Arc<i64> = Arc::new(42);
+	let weak = Arc::downgrade(&arc);
+	let ptr = Arc::into_raw(arc);
+
+	unsafe {
+		Arc::increment_shared_strong_count(ptr);
+		assert!(weak.upgrade().is_ok());
+		Arc::decrement_shared_strong_count(ptr);
+		assert!(weak.upgrade().is_ok());
+		Arc::decrement_shared_strong_count(ptr);
+		assert!(weak.upgrade().is_err());
+	}
+}
+
+
+#[test]
+fn test_cyclic() {
+	struct Cycle (Weak<Cycle>, Test);
+
+	DROP_COUNTER.with(|x| x.set(0));
+
+	{
+		let rc = Rc::new_cyclic(|w| { assert!(w.upgrade().is_err()); Cycle(w.clone(), Test(1)) });
+		let weak = &(*rc).0;
+		let arc = weak.upgrade().unwrap();
+		assert!(Rc::ptr_eq(&rc, &arc));
+		assert_eq!(rc.1.0, 1);
+		assert_eq!(&(*rc).0 as *const _, weak as *const _);
+		assert_eq!(DROP_COUNTER.with(|x| x.get()), 0);
+	}
+
+	assert_eq!(DROP_COUNTER.with(|x| x.get()), 1);
+
+	{
+		let arc = Arc::new_cyclic(|w| { assert!(w.upgrade().is_err()); Cycle(w.clone(), Test(2)) });
+		let weak = &(*arc).0;
+		let arc2 = weak.upgrade().unwrap();
+		assert!(Arc::ptr_eq(&arc, &arc2));
+		assert_eq!(arc.1.0, 2);
+		assert_eq!(&(*arc2).0 as *const _, weak as *const _);
+	}
+
+	assert_eq!(DROP_COUNTER.with(|x| x.get()), 2);
+}
+
+
+#[test]
+#[ignore = "takes a long time in debug build"]
+#[should_panic]
+fn test_local_counter_overflow() {
+	let rc: Rc<i64> = Rc::new(1337);
+	let ptr = Rc::into_raw(rc);
+	for _ in 0..=usize::MAX {
+		unsafe {
+			Rc::increment_local_strong_count(ptr);
+		}
+	}
+}
